;;;;
;;;;	OBJECT.SC
;;;;	(c) Sierra On-Line, Inc, 1989
;;;;
;;;;	Author: Jeff Stephenson
;;;;
;;;;	The Object class: root of the class heirarchy.
;;;;
;;;;	Classes:
;;;;		Object
;;;;		Code



(module# OBJECT)

(include	tsn.sh)



;(class RootObj
;	;;; RootObj is defined in the kernel and contains only that which
;	;;; is necessary for an object to be recognized as such, having no
;	;;; behavior of its own.
;
;	(properties
;		-module-
;			;This is the ID of the module node for the load module containing
;			;the object.  Don't change this property.
;		-size-
;			;This is the number of properties in the object.  Don't change.
;		-dict-
;			;This is the ID of the property dictionary (contained
;			;in the defining class) of an object.  Thus if two
;			;objects have equal -dict- properties, they are both
;			;instances of the same class.  Do not change this property,
;			;or the object will stop working.
;
;		-super-
;			;This is the ID of the object's superclass, and is used to
;			;look up a method which is not defined locally.  Don't change
;			;it, or the object will cease functioning.
;
;		-info-
;			;Bit-mapped information about the object.  Bits are:
;			;	CLONED		0 = object is static
;			;					1 = object was created with new:
;			;	NODISPOSE	0 = dispose: object if CLONED
;			;					1 = don't dispose: object
;			;	NODISPLAY	0 = display object in debugger's object display
;			;					1 = don't display object in debugger's object
;			;						 display
;			;	CLASS			0 = object is an instance of a class
;			;					1 = object is a class
;	)
;)



(class Object kindof RootObj
	;;; Class Object is the superclass of all other Script objects.  It
	;;; defines the behavior which is expected of all objects, ensuring
	;;; that all objects will respond to a certain set of selectors.
	
	(properties
		id		name	"Obj"			;print name of the object
		int	sid	0				;system id of networked object
	)
	
	(methods
		new					;create a new instance of the class/object
		init					;initialize the object
		doit					;do your thing
		dispose				;dispose of this instance
		perform				;execute code in your environment
		isClass				;is the receiver a class or an instance?
		isKindOf				;is object/class a subclass of something?
		isMemberOf			;is object an instance of something?
		respondsTo			;does object responds to a message?
		yourself				;answer your object ID
		show					;show yourself somehow

		mousedOn
		handleMsg			;deal with a remote message
		handleRet			;deal with a ret value from a Dialog
		joinNet				;register with the host
		leaveNet				;kill object on host
		setInt				;set integer properties throughout net
		setStr				;set string property throughout net
		getProp				;get properties of this object
		setSid				;set the SID for this object
		invokeMethod		;invoke a method of this object across the net

		writeTo				;write this object to a file
		readFrom				;read this object from a file
	)
	
	
	(method (new withSid &tmp obj)
		;; Returns the ID of a new instance of the receiver.  This new
		;; instance has the same methods and property values as the
		;; receiver.  The property values can be changed, the methods
		;; cannot.
		
		(= obj (Clone self))
		(if argc
			(self setSid: withSid)
		)
		(return obj)
	)
	
	
	(method (init)
		;; Initializes the receiver.  If the object is to show up on the
		;; screen (Views, Props, Actors and Egos), this is the message
		;; which will make the it show itself.  The default does nothing.
	)
	

	(method (doit)
		;; Do your thing, whatever that may be.  Sometimes this is sent
		;; when the object is selected, sometimes (as for Actors) it is
		;; sent for each animation cycle.  The default does nothing but
		;; return the object ID.
		
		(return self)
	)
	
	
	(method (dispose)
		;; Dispose of the receiver.  If the receiver was created using
		;; new:, this reclaims the memory occupied by that object.  In
		;; this case it is an error to refer to the object ID being
		;; disposed of once the dispose: message has been sent.
		;; Subclasses of Object should make sure that they dispose of
		;; any objects which they created before doing a (super dispose:).
		
		(if sid
			(SID SIDDELETE sid self)
			(self leaveNet:)
		)
		(DisposeClone self)
	)

	
	(method (show)
		;; Displays the receiver in a meaningful way.  This is primarily
		;; used for debugging, in order to find out which objects are
		;; present.  The default is to print the string produced by
		;; showStr:.
		
		(if name
			(Print name)
		else
			(Printf "ID: $%x" self)
		)
	)
	
	
	(method (perform theCode)
		;; Executes the doit: method of theCode with the first parameter
		;; of the method being the object ID of the receiver of perform:.
		;; Any number of parameters may be passed.  Returns the result.
		
		(return (theCode doit: self &rest))
	)
	
	
	(method (respondsTo selector)
		;; Returns TRUE if the receiver has a method corresponding to
		;; selector, FALSE otherwise.
		
		(return (RespondsTo self selector))
	)


	(method (isClass)
		(return (& -info- CLASS))
	)
	
	
	(method (isMemberOf what)
		;; Returns TRUE if the receiver is an instance of what, FALSE
		;; otherwise.
		
		(return
			(and
				(& (what -info-?) CLASS)
				(not (& -info- CLASS))
				(== -dict- (what -dict-?))
			)
		)
	)
	
	
	(method (isKindOf what &tmp theSuper)
		;; Returns TRUE if the receiver is a subclass (or an instance of
		;; a subclass) of class, FALSE otherwise.
		
		(= theSuper (self -super-?))
		(return
			(or
				(== -dict- (what -dict-?))
				(and
					(IsObject theSuper)
					(theSuper isKindOf: what)
				)
			)
		)
	)
	
	
	(method (yourself)
		;; Returns the object ID of the receiver.  This is useful as the
		;; last in a series of messages to an object to force the entire
		;; send to return the ID of the object.
		
		(return self)
	)


	(method (mousedOn)
		(return FALSE)
	)


	(method (handleMsg msg)
		(switch (msg command?)
			(SN_SET
				(msg doit:)
			)
			(SN_SETINT
				(msg doit:)
			)
			(SN_SETSTR
				(msg doit:)
			)
			(SN_OBJID
				(if (== self (msg to?))
					(= sid (msg sid?))
					(SID SIDADD sid self)
				)
			)
			(SN_OBJFREE
				(self dispose:)
			)
			(SN_RMI							;remote method invocation
				(msg doit:)
			)
		)
	)

	(method (handleRet)
	)

	(method (joinNet objNum objType)
		(TSN TSN_SEND
			"wwwbbww"
			SN_OBJNEW
			0
			self
			(if (!= argc 2) NEW_OBJECT else objType)
			(theGame number?)
			objNum
			-size-
		)
	)


	(method (leaveNet)
		(TSN TSN_SEND
			"w+"
			SN_OBJFREE
			sid
			0
		)
	)


	(method (setInt p &tmp i str)
	;; Set various properties of the receiver.

		;Convert the property selectors to offsets within the recipient.
		(for	((= i 0))
				(< i argc)
				((+= i 2))

			(= [p i] (ObjPropOffset self [p i]))
		)

		;Send the request to the host.
		(TSN TSN_SEND
			"w+"
			SN_SETINT
			sid
			sid
			&rest p
		)
	)


	(method (invokeMethod methNum)
		(TSN TSN_SEND
			"w+"
			SN_RMI				;RMI command
			sid					;to
			methNum
			&rest					;method # & arguments
		)
		(self methNum &rest)
	)


	(method (setStr what value)
	;; Set the string property 'what' of the receiver to the string
	;; whose ID is 'value'.
		(TSN TSN_SEND
			"wwwws"
			SN_SETSTR
			sid
			sid
			(ObjPropOffset self what)
			value
		)
	)


	(method (getProp p &tmp i str)
	;; Get various properties of the receiver.

		;Convert the property selectors to offsets within the receiver
		(for	((= i 0))
				(< i argc)
				((++ i))

			(= [p i] (ObjPropOffset self [p i]))
		)

		;Send the request to the host.
		(TSN TSN_SEND
			"w+"
			SN_EGETPROP
			sid
			sid
			&rest p
		)
	)


	(method (setSid newSid)
		(= sid newSid)
		(SID SIDADD newSid self)
	)


	(method (writeTo)
	)


	(method (readFrom)
	)
)



(class Code kindof Object
	;;; Class Code is used to attach code to various objects (as in
	;;; the 'viewer' property of an Actor) or for use in the perform:
	;;; method of Object.  In either case, it is only the doit: method
	;;; of the class which is important.
	

	(method (doit)
		;; This is the code which you want executed.  Calling conventions
		;; will be dictated by the use of this class.
	)
)

